# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2020 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

import json
import logging
from decimal import Decimal
from lxml import etree

from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.orm import joinedload

from hydra_base import db
from hydra_base.db.model import Template, TemplateType, TypeAttr, Network, Dataset, Metadata
from hydra_base.lib.data import add_dataset
from hydra_base.exceptions import HydraError
from hydra_base import config
from hydra_base.util import dataset_util
from hydra_base.lib import units
from hydra_base.util.permissions import required_perms

from hydra_base.lib.template.utils import check_dimension, get_attr_by_name_and_dimension

log = logging.getLogger(__name__)


@required_perms("get_template")
def get_template_as_xml(template_id, **kwargs):
    """
        Turn a template into an xml template
    """
    template_xml = etree.Element("template_definition")

    template_i = db.DBSession.query(Template).filter(
        Template.id == template_id).options(
            joinedload(Template.templatetypes)\
            .joinedload(TemplateType.typeattrs)\
            .joinedload(TypeAttr.default_dataset)\
            .joinedload(Dataset.metadata)
        ).one()

    template_name = etree.SubElement(template_xml, "template_name")
    template_name.text = template_i.name
    template_description = etree.SubElement(template_xml, "template_description")
    template_description.text = template_i.description
    resources = etree.SubElement(template_xml, "resources")

    for type_i in template_i.templatetypes:
        xml_resource = etree.SubElement(resources, "resource")

        resource_type = etree.SubElement(xml_resource, "type")
        resource_type.text = type_i.resource_type

        name = etree.SubElement(xml_resource, "name")
        name.text = type_i.name

        description = etree.SubElement(xml_resource, "description")
        description.text = type_i.description

        alias = etree.SubElement(xml_resource, "alias")
        alias.text = type_i.alias

        if type_i.layout is not None and type_i.layout != "":
            layout = _get_layout_as_etree(type_i.layout)
            xml_resource.append(layout)

        for type_attr in type_i.typeattrs:
            _make_attr_element_from_typeattr(xml_resource, type_attr)

        resources.append(xml_resource)

    xml_string = etree.tostring(template_xml, encoding="unicode")

    return xml_string

@required_perms("add_template")
def import_template_xml(template_xml, allow_update=True, **kwargs):
    """
        Add the template, type and typeattrs described
        in an XML file.

        Delete type, typeattr entries in the DB that are not in the XML file
        The assumption is that they have been deleted and are no longer required.
    """
    user_id = kwargs.get('user_id')

    template_xsd_path = config.get('templates', 'template_xsd_path')
    xmlschema_doc = etree.parse(template_xsd_path)

    xmlschema = etree.XMLSchema(xmlschema_doc)

    xml_tree = etree.fromstring(template_xml)

    xmlschema.assertValid(xml_tree)

    template_name = xml_tree.find('template_name').text
    template_description = xml_tree.find('template_description')
    if template_description is not None:
        template_description = template_description.text

    template_layout = None
    if xml_tree.find('layout') is not None and \
               xml_tree.find('layout').text is not None:
        layout = xml_tree.find('layout')
        layout_string = get_etree_layout_as_dict(layout)
        template_layout = json.dumps(layout_string)

    try:
        tmpl_i = db.DBSession.query(Template).filter(Template.name == template_name)\
            .options(joinedload(Template.templatetypes)\
            .joinedload(TemplateType.typeattrs)\
            .joinedload(TypeAttr.attr)
            ).one()

        if allow_update == False:
            raise HydraError("Existing Template Found with name %s"%(template_name,))
        else:
            log.debug("Existing template found. name=%s", template_name)
            tmpl_i.layout = template_layout
            tmpl_i.description = template_description
    except NoResultFound:
        log.debug("Template not found. Creating new one. name=%s", template_name)
        tmpl_i = Template(name=template_name,
                          description=template_description, layout=template_layout)
        db.DBSession.add(tmpl_i)

    types = xml_tree.find('resources')
    #Delete any types which are in the DB but no longer in the XML file
    type_name_map = {r.name:r.id for r in tmpl_i.templatetypes}
    attr_name_map = {}
    for type_i in tmpl_i.templatetypes:
        for typeattr in type_i.typeattrs:
            attr_name_map[typeattr.attr.name] = (typeattr.attr.id, typeattr.type_id)

    existing_types = set([r.name for r in tmpl_i.templatetypes])

    new_types = set([r.find('name').text for r in types.findall('resource')])

    types_to_delete = existing_types - new_types

    for type_to_delete in types_to_delete:
        type_id = type_name_map[type_to_delete]
        try:
            type_i = db.DBSession.query(TemplateType).filter(TemplateType.id==type_id).one()
            log.debug("Deleting type %s", type_i.name)
            db.DBSession.delete(type_i)
        except NoResultFound:
            pass

    #Add or update types.
    for resource in types.findall('resource'):
        type_name = resource.find('name').text
        #check if the type is already in the DB. If not, create a new one.
        type_is_new = False
        if type_name in existing_types:
            type_id = type_name_map[type_name]
            type_i = db.DBSession.query(TemplateType).filter(
                TemplateType.id == type_id).options(
                    joinedload(TemplateType.typeattrs)\
                    .joinedload(TypeAttr.attr)
                ).one()

        else:
            log.debug("Type %s not found, creating new one.", type_name)
            type_i = TemplateType()
            type_i.name = type_name
            tmpl_i.templatetypes.append(type_i)
            type_is_new = True

        if resource.find('alias') is not None:
            type_i.alias = resource.find('alias').text

        if resource.find('description') is not None:
            type_i.description = resource.find('description').text

        if resource.find('type') is not None:
            type_i.resource_type = resource.find('type').text

        if resource.find('layout') is not None and \
            resource.find('layout').text is not None:
            layout = resource.find('layout')
            layout_string = get_etree_layout_as_dict(layout)
            type_i.layout = json.dumps(layout_string)

        #delete any TypeAttrs which are in the DB but not in the XML file
        existing_attrs = []
        if not type_is_new:
            for r in tmpl_i.templatetypes:
                if r.name == type_name:
                    for typeattr in r.typeattrs:
                        existing_attrs.append(typeattr.attr.name)

        existing_attrs = set(existing_attrs)

        template_attrs = set([r.find('name').text for r in resource.findall('attribute')])

        attrs_to_delete = existing_attrs - template_attrs
        for attr_to_delete in attrs_to_delete:
            attr_id, type_id = attr_name_map[attr_to_delete]
            try:
                attr_i = db.DBSession.query(TypeAttr).filter(
                    TypeAttr.attr_id == attr_id,
                    TypeAttr.type_id == type_id).options(joinedload(TypeAttr.attr)).one()
                db.DBSession.delete(attr_i)
                log.debug("Attr %s in type %s deleted", attr_i.attr.name, attr_i.templatetype.name)
            except NoResultFound:
                log.debug("Attr %s not found in type %s",attr_id, type_id)
                continue

        #Add or update type typeattrs
        for attribute in resource.findall('attribute'):
            new_typeattr = _parse_xml_typeattr(type_i, attribute, user_id=user_id)

    db.DBSession.flush()

    return tmpl_i

@required_perms('get_network')
def get_network_as_xml_template(network_id, **kwargs):
    """
        Turn an existing network into an xml template
        using its attributes.
        If an optional scenario ID is passed in, default
        values will be populated from that scenario.
    """
    template_xml = etree.Element("template_definition")

    net_i = db.DBSession.query(Network).filter(Network.id == network_id).one()

    template_name = etree.SubElement(template_xml, "template_name")
    template_name.text = "TemplateType from Network %s"%(net_i.name)
    layout = _get_layout_as_etree(net_i.layout)

    resources = etree.SubElement(template_xml, "resources")
    if net_i.attributes:
        net_resource = etree.SubElement(resources, "resource")

        resource_type = etree.SubElement(net_resource, "type")
        resource_type.text = "NETWORK"

        resource_name = etree.SubElement(net_resource, "name")
        resource_name.text = net_i.name

        layout = _get_layout_as_etree(net_i.layout)
        if layout is not None:
            net_resource.append(layout)

        for net_attr in net_i.attributes:
            _make_attr_element_from_resourceattr(net_resource, net_attr)

        resources.append(net_resource)

    existing_types = {'NODE': [], 'LINK': [], 'GROUP': []}
    for node_i in net_i.nodes:
        node_attributes = node_i.attributes
        attr_ids = [res_attr.attr_id for res_attr in node_attributes]
        if len(attr_ids) > 0 and attr_ids not in existing_types['NODE']:

            node_resource = etree.Element("resource")

            resource_type = etree.SubElement(node_resource, "type")
            resource_type.text = "NODE"

            resource_name = etree.SubElement(node_resource, "name")
            resource_name.text = node_i.node_name

            layout = _get_layout_as_etree(node_i.layout)

            if layout is not None:
                node_resource.append(layout)

            for node_attr in node_attributes:
                _make_attr_element_from_resourceattr(node_resource, node_attr)

            existing_types['NODE'].append(attr_ids)
            resources.append(node_resource)

    for link_i in net_i.links:
        link_attributes = link_i.attributes
        attr_ids = [link_attr.attr_id for link_attr in link_attributes]
        if len(attr_ids) > 0 and attr_ids not in existing_types['LINK']:
            link_resource = etree.Element("resource")

            resource_type = etree.SubElement(link_resource, "type")
            resource_type.text = "LINK"

            resource_name = etree.SubElement(link_resource, "name")
            resource_name.text = link_i.link_name

            layout = _get_layout_as_etree(link_i.layout)

            if layout is not None:
                link_resource.append(layout)

            for link_attr in link_attributes:
                _make_attr_element_from_resourceattr(link_resource, link_attr)

            existing_types['LINK'].append(attr_ids)
            resources.append(link_resource)

    for group_i in net_i.resourcegroups:
        group_attributes = group_i.attributes
        attr_ids = [group_attr.attr_id for group_attr in group_attributes]
        if len(attr_ids) > 0 and attr_ids not in existing_types['GROUP']:
            group_resource = etree.Element("resource")

            resource_type = etree.SubElement(group_resource, "type")
            resource_type.text = "GROUP"

            resource_name = etree.SubElement(group_resource, "name")
            resource_name.text = group_i.group_name


            for group_attr in group_attributes:
                _make_attr_element_from_resourceattr(group_resource, group_attr)

            existing_types['GROUP'].append(attr_ids)
            resources.append(group_resource)

    xml_string = etree.tostring(template_xml, encoding="unicode")

    return xml_string

def _parse_xml_attribute(attribute):
    """
        Parse an attribute as defined in the template XML file
    """
    dimension_i = None

    attribute_name = attribute.find('name').text.strip()

    if attribute.find('dimension') is not None:
        dimension_name = attribute.find('dimension').text

        if dimension_name is not None and dimension_name.strip() != '':
            dimension_i = units.get_dimension_by_name(dimension_name.strip())

    elif attribute.find('unit') is not None:
        # Found the unit
        unit_abbr = attribute.find('unit').text
        if unit_abbr is not None and unit_abbr.strip() != '':
            unit_id = units.get_unit_by_abbreviation(unit_abbr).id
            dimension_i = units.get_dimension_by_unit_id(unit_id,
                                                         do_accept_unit_id_none=True)

    if dimension_i is None:
        attr = get_attr_by_name_and_dimension(attribute_name, None)
    else:
        attr = get_attr_by_name_and_dimension(attribute_name, dimension_i.id)

    db.DBSession.flush()

    return attr

def _parse_xml_typeattr(type_i, attribute, user_id=None):
    """
        convert a typeattr etree element and turn it into a hydra type attr
    """

    attr = _parse_xml_attribute(attribute)

    for ta in type_i.typeattrs:
        if ta.attr_id == attr.id:
            # Find the TypeAttr
            typeattr_i = ta
            break
    else:
        # Creating a new TypeAttr
        typeattr_i = TypeAttr()
        log.debug("Creating type attr: type_id=%s, attr_id=%s", type_i.id, attr.id)
        typeattr_i.type_id = type_i.id
        typeattr_i.attr_id = attr.id
        type_i.typeattrs.append(typeattr_i)
        db.DBSession.add(typeattr_i)

    typeattr_unit_id = None
    if attribute.find('unit') is not None:
        # Found the unit as child at first level
        unit = attribute.find('unit').text
        if unit not in ('', None):
            typeattr_unit_id = units.get_unit_by_abbreviation(unit).id

    if typeattr_unit_id is not None:
        typeattr_i.unit_id = typeattr_unit_id

    check_dimension(typeattr_i)

    if attribute.find('description') is not None:
        typeattr_i.description = attribute.find('description').text

    if attribute.find('properties') is not None:
        properties_string = get_etree_layout_as_dict(attribute.find('properties'))
        typeattr_i.properties = str(properties_string)

    if attribute.find('is_var') is not None:
        typeattr_i.attr_is_var = attribute.find('is_var').text
    if attribute.find('data_type') is not None:
        typeattr_i.data_type = attribute.find('data_type').text

    # Analyzing the "default" node
    if attribute.find('default') is not None:
        default = attribute.find('default')

        dataset_unit_id = None
        if default.find('unit') is not None:
            dataset_unit = default.find('unit').text
            if dataset_unit not in ('', None):
                dataset_unit_id = units.get_unit_by_abbreviation(dataset_unit).id

        if dataset_unit_id is None and typeattr_i.unit_id is not None:
            dataset_unit = typeattr_i.unit_id

        if dataset_unit_id is not None and typeattr_i.unit_id is not None:
            if dataset_unit_id != typeattr_i.unit_id:
                raise HydraError(f"Default value has a unit of {typeattr_i.unit_id}"+
                                 "but the attribute"+
                                 f" says the unit should be: {dataset_unit_id}")

        val = default.find('value').text
        try:
            Decimal(val)
            data_type = 'scalar'
        except:
            data_type = 'descriptor'

        dataset = add_dataset(data_type,
                              val,
                              dataset_unit_id,
                              name="%s Default"%attr.name,
                              user_id=user_id)

        typeattr_i.default_dataset_id = dataset.id

    if attribute.find('restrictions') is not None:
        restriction = str(dataset_util.get_restriction_as_dict(attribute.find('restrictions')))
        typeattr_i.data_restriction = restriction
    else:
        typeattr_i.data_restriction = None


    return typeattr_i

def _make_attr_element_from_typeattr(parent, type_attr_i):
    """
        General function to add an attribute element to a resource element.
        resource_attr_i can also e a type_attr if being called from get_tempalte_as_xml
    """

    attr = _make_attr_element(parent, type_attr_i.attr)

    if type_attr_i.unit_id is not None:
        attr_unit = etree.SubElement(attr, 'unit')
        attr_unit.text = units.get_unit(type_attr_i.unit_id).abbreviation

    attr_is_var = etree.SubElement(attr, 'is_var')
    attr_is_var.text = type_attr_i.attr_is_var

    if type_attr_i.data_type is not None:
        attr_data_type = etree.SubElement(attr, 'data_type')
        attr_data_type.text = type_attr_i.data_type

    if type_attr_i.data_restriction is not None:
        attr_data_restriction = etree.SubElement(attr, 'restrictions')
        attr_data_restriction.text = type_attr_i.data_restriction

    return attr

def _make_attr_element_from_resourceattr(parent, resource_attr_i):
    """
        General function to add an attribute element to a resource element.
    """

    attr = _make_attr_element(parent, resource_attr_i.attr)

    attr_is_var = etree.SubElement(attr, 'is_var')
    attr_is_var.text = resource_attr_i.attr_is_var

    return attr

def _make_attr_element(parent, attr_i):
    """
        create an attribute element from an attribute DB object
    """
    attr = etree.SubElement(parent, "attribute")

    attr_name = etree.SubElement(attr, 'name')
    attr_name.text = attr_i.name

    attr_desc = etree.SubElement(attr, 'description')
    attr_desc.text = attr_i.description

    attr_dimension = etree.SubElement(attr, 'dimension')
    attr_dimension.text = units.get_dimension(attr_i.dimension_id,
                                              do_accept_dimension_id_none=True).name

    return attr

def get_etree_layout_as_dict(layout_tree):
    """
    Convert something that looks like this:
    <layout>
        <item>
            <name>color</name>
            <value>red</value>
        </item>
        <item>
            <name>shapefile</name>
            <value>blah.shp</value>
        </item>
    </layout>
    Into something that looks like this:
    {
        'color' : ['red'],
        'shapefile' : ['blah.shp']
    }
    """
    layout_dict = dict()

    for item in layout_tree.findall('item'):
        name = item.find('name').text
        val_element = item.find('value')
        value = val_element.text.strip()
        if value == '':
            children = val_element.getchildren()
            value = etree.tostring(children[0], pretty_print=True, encoding="unicode")
        layout_dict[name] = value
    return layout_dict

def _get_layout_as_etree(layout_dict):
    """
    Convert something that looks like this:
    {
        'color' : ['red'],
        'shapefile' : ['blah.shp']
    }

    Into something that looks like this:
    <layout>
        <item>
            <name>color</name>
            <value>red</value>
        </item>
        <item>
            <name>shapefile</name>
            <value>blah.shp</value>
        </item>
    </layout>
    """
    if layout_dict is None:
        return None

    layout = etree.Element("layout")
    layout_dict = eval(layout_dict)
    for k, v in layout_dict.items():
        item = etree.SubElement(layout, "item")
        name = etree.SubElement(item, "name")
        name.text = k
        value = etree.SubElement(item, "value")
        value.text = str(v)

    return layout



# (c) Copyright 2013, 2014, University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#
#!/usr/bin/env python
# -*- coding: utf-8 -*-

import copy
import datetime
import json
import pytest

from sqlalchemy.exc import IntegrityError

import hydra_base
from hydra_base.exceptions import ResourceNotFoundError
from hydra_base.exceptions import HydraError, PermissionError
from hydra_base.lib.objects import JSONObject, Dataset
from hydra_base.util.hydra_dateutil import timestamp_to_ordinal

from .templates.test_templates import template, template_json_object

import logging
log = logging.getLogger(__name__)

class TestRules:

    def add_rule(self, client, target=None, name='A Test Rule', text='e=mc^2',
                 ref_key=None, ref_id=None, types=[]):
        """
            A utility function which creates a rule and associates it
            with either a resource type, resource instance or resource
            instance / scenario pair.
        """

        rule = JSONObject({
            'name'        : name,
            'value'       : text,
            'ref_key'     : "NETWORK" if ref_key is None else ref_key,
            'ref_id'      : target.id if ref_id is None else ref_id,
            'types'       : [{'code':typecode} for typecode in types]
        })

        c_rule = client.add_rule(rule=rule)
        new_rule_j = JSONObject(c_rule)
        return new_rule_j

    def test_add_rule_type_definition(self, client):
        client.add_rule_type_definition(JSONObject({'name':'A new Rule', 'code':'a_new_rule'}))

        rule_type_j = client.get_rule_type_definition('a_new_rule')

        assert rule_type_j.code == 'a_new_rule'

    def test_get_rule_type_definitions(self, client):
        client.add_rule_type_definition(JSONObject({'name':'A new Rule', 'code':'a_new_rule'}))
        client.add_rule_type_definition(JSONObject({'name':'A new Rule 1', 'code':'a_new_rule_1'}))

        rule_types_j = client.get_rule_type_definitions()

        assert len(rule_types_j) == 2

    def test_get_rule_type_definition(self, client):
        client.add_rule_type_definition(JSONObject({'name':'A new Rule', 'code':'a_new_rule'}))

        rule_type_j = client.get_rule_type_definition('a_new_rule')

        assert rule_type_j.code == 'a_new_rule'

    def test_get_rule_by_id(self, client, network_with_data):

        new_rule_j = self.add_rule(client, network_with_data)

        rule_j = JSONObject(client.get_rule(new_rule_j.id))

        assert rule_j.name == 'A Test Rule'
        assert rule_j.value == 'e=mc^2'

        #2nd error handle is to cover errors coming from the server
        with pytest.raises((ResourceNotFoundError, HydraError)):
            rule_j = JSONObject(client.get_rule(2))

        #temporarily set the client user_id to a different (non-admin) user
        client.user_id = 5 #5 is not an admin and is not an owner of the rule
        with pytest.raises((PermissionError, HydraError)):
            client.get_rule(new_rule_j.id)

    def test_clone_network_with_rules(self, client, network_with_data):

        net_rule_j = self.add_rule(client, network_with_data)

        cloned_network_id = client.clone_network(network_with_data.id)

        cloned_network = client.get_network(cloned_network_id)

        network_rules = client.get_resource_rules('NETWORK', cloned_network.id)
        assert len(network_rules) == 1
        assert network_rules[0].value == net_rule_j.value;

    def test_share_network_with_rules(self, client, network_with_data):

        net_rule_j = self.add_rule(client, network_with_data)

        cloned_network_id = client.clone_network(network_with_data.id)

        cloned_network = client.get_network(cloned_network_id)
        network_rules = client.get_resource_rules('NETWORK', cloned_network.id)

        assert len(network_rules) == 1
        assert network_rules[0].value == net_rule_j.value

        client.share_network(cloned_network.id, ['UserC'], 'Y', 'Y')
        client.user_id = pytest.user_c.id
        client.get_network(cloned_network_id)

        #User c should see the same rules
        user_c_rules = client.get_resource_rules('NETWORK', cloned_network.id)
        assert len(user_c_rules) == 1
        assert user_c_rules[0].value == net_rule_j.value

    def test_get_rules_by_type(self, client, network_with_data):
        ruletype_A_j = client.add_rule_type_definition(JSONObject({'name':'A new Rule',
                                                                   'code':'a_new_rule'}))
        ruletype_B_j = client.add_rule_type_definition(JSONObject({'name':'A new Rule 1',
                                                                   'code':'a_new_rule_1'}))

        scenario_id = network_with_data.scenarios[0].id

        #Create 3 rules, 2 of type A and 1 of type B
        self.add_rule(client, network_with_data, name="Test1", types=[ruletype_A_j.code])
        self.add_rule(client, network_with_data, name="Test2", types=[ruletype_A_j.code])
        self.add_rule(client, network_with_data, name="Test3", types=[ruletype_A_j.code, ruletype_B_j.code])
        self.add_rule(client, network_with_data, name="Rule Type B", types=[ruletype_B_j.code])

        #Get all the rules of type A, of which there should be 2
        rules_of_type = client.get_rules_of_type(ruletype_A_j.code)

        assert len(rules_of_type) == 3

    def test_add_rule_type(self, client, network_with_data):
        ruletype_A_j = client.add_rule_type_definition(JSONObject({'name':'A new Rule', 'code':'a_new_rule'}))
        ruletype_B_j = client.add_rule_type_definition(JSONObject({'name':'A new Rule 1', 'code':'a_new_rule_1'}))

        scenario_id = network_with_data.scenarios[0].id

        #Create 3 rules, 2 of type A and 1 of type B
        self.add_rule(client, network_with_data, name="Test1", types=[ruletype_A_j.code])
        self.add_rule(client, network_with_data, name="Test2", types=[ruletype_A_j.code])
        rule3 = self.add_rule(client, network_with_data, name="Test3", types=[ruletype_A_j.code])

        #Get all the rules of type A, of which there should be 2
        rules_of_type = client.get_rules_of_type(ruletype_A_j.code)

        assert len(rules_of_type) >= 3
        #CHeck that the added nodes are indeed present
        assert {'Test1', 'Test2', 'Test3'}.issubset({r.name for r in rules_of_type})

        client.set_rule_type(rule3.id, ruletype_B_j.code)

        rules_of_type_b = client.get_rules_of_type(ruletype_B_j.code)
        assert len(rules_of_type_b) >= 1
        assert 'Test3' in [r.name for r in rules_of_type_b]

    def test_add_rule1(self, client, network_with_data):
        rule_network = client.get_network(network_with_data.id)

        #Sharae the network with user A to thest the sharing feature.
        client.share_network(network_with_data.id, ['UserA'], 'Y', 'Y')

        rulename = 'Added Rule'
        ruletext = 'e=mc^3'#yes this is delibrate, so it's different to the default
        new_rule_j = self.add_rule(client, network_with_data, name=rulename, text=ruletext)

        assert new_rule_j.id is not None
        assert new_rule_j.name == rulename
        assert new_rule_j.value == ruletext

        # Rule names must be unique within a Network
        with pytest.raises(IntegrityError):
            duplicate_rule = self.add_rule(client, network_with_data, name=rulename, text=ruletext)

        #sanity check to ensure we're actually testing that the ownership functionality
        #is testing correctly
        rule_network = client.get_network(network_with_data.id)
        assert len(rule_network.owners) == len(new_rule_j.owners)


    def test_clone_rule(self, client, network_with_data, projectmaker):

        ruletype_A_j = client.add_rule_type_definition(JSONObject({'name':'A new Rule', 'code':'a_new_rule'}))
        ruletype_B_j = client.add_rule_type_definition(JSONObject({'name':'A new Rule 1', 'code':'a_new_rule_1'}))

        rulename = 'Added Rule'
        ruletext = 'e=mc^3'#yes this is delibrate, so it's different to the default
        new_rule_j = self.add_rule(client,
                                   target=network_with_data,
                                   name=rulename,
                                   text=ruletext,
                                   types=[ruletype_A_j.code, ruletype_B_j.code])

        assert new_rule_j.id is not None
        assert new_rule_j.name == rulename
        assert new_rule_j.value == ruletext

        project = projectmaker.create('Target Network Parent')
        network = JSONObject()
        network.name = f"New network {datetime.datetime.now()}"
        network.description = "Clone rule target network"
        network.project_id = project.id

        target_network = client.add_network(network)

        # Clone to rule to new network
        cloned_rule = client.clone_rule(new_rule_j.id, target_ref_key="NETWORK", target_ref_id=target_network.id)

        assert cloned_rule.name == new_rule_j.name
        assert cloned_rule.value == new_rule_j.value
        assert cloned_rule.id is not None
        assert cloned_rule.id != new_rule_j.id
        assert len(cloned_rule.types) == 2
        assert [t.code for t in new_rule_j.types] == [t.code for t in cloned_rule.types]

        # Clone rule to new project
        src_proj = projectmaker.create("Rule Source Project")
        dest_proj = projectmaker.create("Rule Destination Project")


        proj_rulename = "Project Rule"
        proj_ruletext = "e=mc^3"
        proj_rule = self.add_rule(client,
                                  name=proj_rulename,
                                  text=proj_ruletext,
                                  types=[ruletype_A_j.code],
                                  ref_key="PROJECT",
                                  ref_id=project.id)

        proj_cloned_rule = client.clone_rule(
                             proj_rule.id,
                             target_ref_key="PROJECT",
                             target_ref_id=dest_proj.id)

        assert proj_cloned_rule.name == proj_rule.name
        assert proj_cloned_rule.value == proj_rule.value
        assert proj_cloned_rule.id is not None
        assert proj_cloned_rule.id != proj_rule.id
        assert len(proj_cloned_rule.types) == 1


    def test_project_rule(self, client, projectmaker):
        project = projectmaker.create("Rule Parent Project")
        ruletype_A_j = client.add_rule_type_definition(JSONObject({'name':'A new Rule', 'code':'a_new_rule'}))
        rulename = "Rule 001"
        ruletext = "int(1)"
        new_rule_j = self.add_rule(client,
                                   target=project,
                                   ref_key="PROJECT",
                                   ref_id=project.id,
                                   name=rulename,
                                   text=ruletext,
                                   types=[ruletype_A_j.code])

        assert new_rule_j.id is not None
        assert new_rule_j.name == rulename
        assert new_rule_j.value == ruletext

        # Rule names must be unique within a Project
        with pytest.raises(IntegrityError):
            duplicate_rule = self.add_rule(client,
                                           target=project,
                                           ref_key="PROJECT",
                                           ref_id=project.id,
                                           name=rulename,
                                           text=ruletext,
                                           types=[ruletype_A_j.code])

        ret_rules = client.get_project_rules(project.id)

        assert ret_rules[0].id == new_rule_j.id
        assert ret_rules[0].name == new_rule_j.name
        assert ret_rules[0].value == new_rule_j.value

        res_ret_rules = client.get_resource_rules("PROJECT", project.id)

        assert res_ret_rules[0].id == new_rule_j.id
        assert res_ret_rules[0].name == new_rule_j.name
        assert res_ret_rules[0].value == new_rule_j.value

    def test_update_rule(self, client, network_with_data):

        typecode = 'a_new_rule_type'
        typecode1 = 'a_new_rule_type_1'

        client.add_rule_type_definition(JSONObject({'name':'A new Rule', 'code':typecode}))
        client.add_rule_type_definition(JSONObject({'name':'A new Rule', 'code':typecode1}))

        rulename = 'Added Rule'
        ruletext = 'e=mc^3'#yes this is delibrate, so it's different to the default

        new_rule_j = self.add_rule(client, network_with_data, name=rulename, text=ruletext, types=[typecode])

        new_rule_j = client.get_rule(new_rule_j.id) #do this to get all the DB server defaults

        assert len(new_rule_j.types) == 1

        new_rule_j.name = 'Updated Rule'
        new_rule_j.format = 'text'
        new_rule_j.value    = 'e=mc2' #fix the error
        new_rule_j.types.append(JSONObject({'code':typecode1}))

        client.update_rule(new_rule_j)

        updated_rule_j = client.get_rule(new_rule_j.id)

        assert updated_rule_j.name == 'Updated Rule'
        assert updated_rule_j.value == 'e=mc2'
        assert updated_rule_j.format == 'text'
        assert updated_rule_j.scenario_id is None
        assert len(updated_rule_j.types) == 2


    def test_add_rule_with_type(self, client, network_with_data):

        typecode = 'a_new_rule_type'
        typecode1 = 'a_new_rule_type_1'

        client.add_rule_type_definition(JSONObject({'name':'A new Rule', 'code':typecode}))
        client.add_rule_type_definition(JSONObject({'name':'A new Rule', 'code':typecode1}))

        rulename = 'Added Rule'
        ruletext = 'e=mc^3'#yes this is delibrate, so it's different to the default
        new_rule_j = self.add_rule(client, network_with_data, name=rulename, text=ruletext, types=[typecode, typecode1])

        assert new_rule_j.id is not None
        assert new_rule_j.name == rulename
        assert new_rule_j.value == ruletext
        assert len(new_rule_j.types) == 2


        assert new_rule_j.id in [r.id for r in client.get_rules_of_type(typecode)]

    def test_delete_rule(self, client, network_with_data):

        new_rule_j = self.add_rule(client, network_with_data)

        #prove it exists. If not, it would throw an exception
        client.get_rule(new_rule_j.id)

        assert new_rule_j.id in [r.id for r in client.get_resource_rules('NETWORK', network_with_data.id)]

        client.delete_rule(new_rule_j.id)

        assert new_rule_j.id not in [r.id for r in client.get_resource_rules('NETWORK', network_with_data.id)]

    def test_activate_rule(self, client, network_with_data):

        new_rule_j = self.add_rule(client, network_with_data)

        #prove it exists. If not, it would throw an exception
        client.get_rule(new_rule_j.id)

        assert new_rule_j.id in [r.id for r in client.get_resource_rules('NETWORK', network_with_data.id)]

        client.delete_rule(new_rule_j.id)

        assert new_rule_j.id not in [r.id for r in client.get_resource_rules('NETWORK', network_with_data.id)]

        client.activate_rule(new_rule_j.id)

        #prove it exists again. If not, it would throw an exception
        client.get_rule(new_rule_j.id)

        assert new_rule_j.id in [r.id for r in client.get_resource_rules('NETWORK', network_with_data.id)]


    def test_purge_rule(self, client, network_with_data):
        new_rule_j = self.add_rule(client, network_with_data)

        #prove it exists. If not, it would throw an exception
        client.get_rule(new_rule_j.id)

        assert new_rule_j.id in [r.id for r in client.get_resource_rules('NETWORK', network_with_data.id)]

        client.purge_rule(new_rule_j.id)

        with pytest.raises(HydraError):
            client.get_rule(new_rule_j.id)

    def test_delete_rule_type(self, client, network_with_data):
        ruletype_A_j = client.add_rule_type_definition(JSONObject({'name':'A new Rule', 'code':'a_new_rule'}))
        ruletype_B_j = client.add_rule_type_definition(JSONObject({'name':'A new Rule 1', 'code':'a_new_rule_1'}))

        scenario_id = network_with_data.scenarios[0].id

        #Create 3 rules, 2 of type A and 1 of type B
        self.add_rule(client, network_with_data, name="Test1", types=[ruletype_A_j.code])
        self.add_rule(client, network_with_data, name="Test2", types=[ruletype_A_j.code])
        self.add_rule(client, network_with_data, name="Rule Type B", types=[ruletype_B_j.code])

        #Get all the rules of type A, of which there should be 2
        rules_of_type = client.get_rules_of_type(ruletype_A_j.code)

        assert len(rules_of_type) >= 2
        #Check that the added nodes are indeed present
        assert {'Test1', 'Test2'}.issubset({r.name for r in rules_of_type})

        client.purge_rule_type_definition('a_new_rule')

        assert len(client.get_rules_of_type(ruletype_A_j.code)) == 0

        #check all the rules are still there
        assert len(client.get_resource_rules('NETWORK', network_with_data.id)) == 3

    def test_template_rules(self, client, template_json_object):
        rulename = "Rule 001"
        ruletext = "int(1)"
        new_rule = self.add_rule(client,
                                 target=template_json_object,
                                 ref_key="TEMPLATE",
                                 ref_id=template_json_object.id,
                                 name=rulename,
                                 text=ruletext)

        # Has Rule been added and returned correctly?
        assert new_rule.id is not None
        assert new_rule.name == rulename
        assert new_rule.value == ruletext

        # Rule names must be unique within a Template
        with pytest.raises(IntegrityError):
            duplicate_rule = self.add_rule(client,
                                           target=template_json_object,
                                           ref_key="TEMPLATE",
                                           ref_id=template_json_object.id,
                                           name=rulename,
                                           text=ruletext)

        # Is Rule added to and retrievable from correct Template?
        ret_rules = client.get_template_rules(template_json_object.id)

        assert len(ret_rules) == 1
        assert ret_rules[0].id == new_rule.id
        assert ret_rules[0].name == new_rule.name
        assert ret_rules[0].value == new_rule.value

    def test_rule_permissions(self, client, network_with_data):
        non_admin_user_id = 5
        admin_user_id = 1
        rule = self.add_rule(client, network_with_data)

        # Rule can be retrieved by admin
        ret_rule = client.get_rule(rule.id)

        # But not by non-admin user
        client.user_id = non_admin_user_id
        roles = client.get_user_roles(non_admin_user_id)
        for role in roles:
            assert role.code != "admin"
        with pytest.raises((PermissionError, HydraError)):
            no_read_rule = client.get_rule(rule.id)

        non_admin_rule = self.add_rule(client, network_with_data, name="non_admin_rule")

        # However admin user can read non-admin rule
        client.user_id = admin_user_id
        roles = client.get_user_roles(admin_user_id)
        rolecode_set = set(r.code for r in roles)
        assert "admin" in rolecode_set
        admin_read_rule = client.get_rule(non_admin_rule.id)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

import json
import enum
import logging
import six

from bson.objectid import ObjectId
from bson.errors import InvalidId
from datetime import datetime

from hydra_base.lib.storage import MongoStorageAdapter

from ..exceptions import HydraError
from ..util import (
    generate_data_hash,
    get_json_as_dict,
    get_json_as_string
)

from .HydraTypes.Registry import HydraObjectFactory


log = logging.getLogger(__name__)
mongo = MongoStorageAdapter()
VALID_JSON_FIRST_CHARS = ['{', '[']


class JSONObject(dict):
    """
        A dictionary object whose attributes can be accesed via a '.'.
        Pass in a nested dictionary, a SQLAlchemy object or a JSON string.
    """
    def __init__(self, obj_dict={}, parent=None, extras={}, normalize=True):

        if normalize:
            obj = self.normalise_input(obj_dict)
        else:
            obj = obj_dict

        for k in obj:
            v = obj[k]

            # Preserve tuples as keys of dictionaries
            if isinstance(k, tuple):
                self[k] = v
                continue

            if k == "value_ref":
                continue

            if isinstance(v, JSONObject):
                self[k] = v
            elif k == 'layout':
                #Layout is often valid JSON, but we dont want to treat it as a JSON object necessarily
                dict_layout = get_json_as_dict(v)
                self[k] = dict_layout
            elif isinstance(v, dict):
                #TODO what is a better way to identify a dataset?
                if 'unit_id' in v or 'unit' in v or 'metadata' in v or 'type' in v:
                    self[k] = Dataset(v, obj_dict)
                #The value on a dataset should remain untouched
                elif k == 'value':
                    self[k] = v
                else:
                    self[k] = JSONObject(v, obj_dict, normalize=normalize)
            elif isinstance(v, list):
                #another special case for datasets, to convert a metadata list into a dict
                if k == 'metadata' and obj_dict is not None:
                    if hasattr(obj_dict, 'get_metadata_as_dict'):
                        self[k] = JSONObject(obj_dict.get_metadata_as_dict())
                    else:
                        metadata_dict = JSONObject()
                        if hasattr(obj_dict, 'get'):#special case for resource data and row proxies
                            for m in obj_dict.get('metadata', []):
                                metadata_dict[m.key] = m.value
                        self[k] = metadata_dict

                else:
                    is_list_of_objects = True
                    if len(v) > 0:
                        if isinstance(v[0], (float, int)):
                            is_list_of_objects = False
                        elif isinstance(v[0], six.string_types) and len(v[0]) == 0:
                            is_list_of_objects = False
                        elif isinstance(v[0], six.string_types) and v[0][0] not in VALID_JSON_FIRST_CHARS:
                            is_list_of_objects=False

                    if is_list_of_objects is True:
                        l = [JSONObject(item, obj_dict) for item in v]
                    else:
                        l = v

                    self[k] = l
            #Special case for SQLAlchemy objects, to stop them recursing up and down
            elif hasattr(v, '_sa_instance_state')\
                    and v._sa_instance_state is not None\
                    and v != parent\
                    and hasattr(obj_dict, '_parents')\
                    and obj_dict._parents is not None\
                    and v.__tablename__ not in obj_dict._parents:
                if v.__tablename__.lower() == 'tdataset':
                    l = Dataset(v, obj_dict)
                else:
                    l = JSONObject(v, obj_dict)
                self[k] = l
            #Special case for SQLAlchemy objects, to stop them recursing up and down
            elif hasattr(v, '_sa_instance_state')\
                    and v._sa_instance_state is not None\
                    and v != parent\
                    and hasattr(obj_dict, '_parents')\
                    and obj_dict._parents is not None\
                    and v.__tablename__ in obj_dict._parents:
                continue
            elif isinstance(v, enum.Enum):
                self[k] = v.value
            else:

                if k == '_sa_instance_state':
                    continue

                if parent is not None and type(v) == type(parent):
                    continue

                if isinstance(v, str) and v.replace('.', '', 1).isdigit():
                    v = float(v) if '.' in v else int(v)

                try:
                    if not isinstance(v, int):
                        v = float(v)
                except:
                    pass

                if isinstance(v, datetime):
                    v = six.text_type(v)

                self[six.text_type(k)] = v


        for k, v in extras.items():
            self[k] = v

    def normalise_input(self, obj_dict):
        """
            Pre-process the input dict to ensure that it is compatible with a JSONObject
        """
        asdict_fn = getattr(obj_dict, "asdict", None)
        if asdict_fn is not None and callable(asdict_fn):
            rd = obj_dict.asdict()
            for k, v in rd.items():
                self[k] = v

        if isinstance(obj_dict, str):
            try:
                obj = json.loads(obj_dict)
                assert isinstance(obj, dict), "JSON string does not evaluate to a dict"
            except (AssertionError, json.decoder.JSONDecodeError) as e:
                log.critical("Error with value: %s" , obj_dict)
                log.critical(parent)
                raise ValueError("Unable to read string value. Make sure it's JSON serialisable") from e
        elif hasattr(obj_dict, '_asdict') and obj_dict._asdict is not None:
            """
            The argument is a SQLAlchemy object. This originated from a
            Class.column query so there was no instance to trigger the
            value descriptor's __get__ and the external lookup must be
            performed here.
            """
            obj = obj_dict._asdict()
            if obj.get("value") is not None:
                try:
                    """
                    ref_key may be not None but also not a valid oid string, so
                    must handle InvalidId from ObjectId and possible TypeError
                    if oid inst is created but then matches no document.
                    """
                    oid = ObjectId(obj["value"])
                    doc = mongo.get_document_by_oid_inst(oid)
                    obj["value"] = doc["value"]
                except (TypeError, InvalidId):
                    """ The value wasn't an valid ObjectID, keep the current value """
                    pass
        elif hasattr(obj_dict, '__dict__') and len(obj_dict.__dict__) > 0:
            obj = obj_dict.__dict__
            """
            Handle indirect references.
            The sqlalchemy attr "value_ref" is in the instance __dict__
            but the "value" descriptor class attr is not.
            The "value_ref" must remain present in the __dict__ for
            later external db lookup, but should not be present in the
            returned object whereas the "value" should.
            """
            if "value_ref" in obj:
                if obj_dict.value:
                    obj["value"] = obj_dict.value
        elif isinstance(obj_dict, dict):
            """
            The argument is a dict of uncertain provenance. This can
            originate from SQLAlchemy row._asdict() so must be
            handled similarly.
            """
            obj = obj_dict
            ref_key = obj.get("value")
            if ref_key:
                try:
                    oid = ObjectId(ref_key)
                    doc = mongo.get_document_by_oid_inst(oid)
                    obj["value"] = doc["value"]
                except (TypeError, InvalidId):
                    """ The value wasn't an valid ObjectID, keep the current value """
                    pass
        else:
            #last chance...try to cast it as a dict. Do this for sqlalchemy result proxies.
            try:
                obj = dict(obj_dict)
            except:
                log.critical("Error with value: %s" , obj_dict)
                raise ValueError("Unrecognised value. It must be a valid JSON dict, a SQLAlchemy result or a dictionary.")
        return obj

    def __getattr__(self, name):
        # Make sure that "special" methods are returned as before.

        # Keys that start and end with "__" won't be retrievable via attributes
        if name == '__table__':#special case for SQLAlchemy objects
            return self.get('__table__')
        elif name.startswith('__') and name.endswith('__'):
            return super(JSONObject, self).__getattr__(name)
        else:
            return self.get(name, None)

    def __setattr__(self, key, value):
        self[key] = value

    def as_json(self):

        return json.dumps(self)

    def get_layout(self):
        """
            Return the 'layout' attribute as a json string
            this is a shorcut for backward compatibility.
            calls the `get_json("layout")` function internally

        """
        return self.get_json('layout')

    def get_json(self, key):
        """
            General function to take an attribute of the object, such as
            layout or app data, which is expected to be in JSON format, and
            return it as a JSON blob,

        """
        if self.get(key) is not None:
            return get_json_as_string(self[key])
        else:
            return None

    #Only for type attrs. How best to generalise this?
    def get_properties(self):
        if self.get('properties') and self.get('properties') is not None:
            return six.text_type(self.properties)
        else:
            return None

class ResourceScenario(JSONObject):
    def __init__(self, rs):
        super(ResourceScenario, self).__init__(rs)
        for k, v in rs.items():
            if k == 'dataset':
                self[k] = Dataset(v)


class Dataset(JSONObject):

    def __init__(self, dataset={}, parent=None, extras={}):

        super(Dataset, self).__init__(dataset, parent=parent, extras=extras)


    def __getattr__(self, name):

        # Keys that start and end with "__" won't be retrievable via attributes
        if name.startswith('__') and name.endswith('__'):
            return super(JSONObject, self).__getattr__(name)
        else:
            return self.get(name, None)

    def __setattr__(self, name, value):
        if name == 'value' and value is not None:
            value = six.text_type(value)
        super(Dataset, self).__setattr__(name, value)

    def get_value(self):
        """
            This function is here to match the equivalent one on the tDataset class in model.py
            so that the get_value function can be used without throwing an exception
        """
        return self.value

    def parse_value(self):
        """
            Turn the value of an incoming dataset into a hydra-friendly value.
        """
        try:
            if self.value is None:
                log.warning("Cannot parse dataset. No value specified.")
                return None

            # attr_data.value is a dictionary but the keys have namespaces which must be stripped
            data = six.text_type(self.value)

            if data.upper().strip() in ("NULL", ""):
                return "NULL"

            data = data[0:100]
            log.debug("[Dataset.parse_value] Parsing %s (%s)", data, type(data))

            return HydraObjectFactory.valueFromDataset(self.type, self.value, self.get_metadata_as_dict())

        except Exception as e:
            log.exception(e)
            raise HydraError("Error parsing value %s: %s"%(self.value, e))

    def get_metadata_as_dict(self, user_id=None, source=None):
        """
        Convert a metadata json string into a dictionary.

        Args:
            user_id (int): Optional: Insert user_id into the metadata if specified
            source (string): Optional: Insert source (the name of the app typically) into the metadata if necessary.

        Returns:
            dict: THe metadata as a python dictionary
        """

        if self.metadata is None or self.metadata == "":
            return {}

        metadata_dict = self.metadata if isinstance(self.metadata, dict) else json.loads(self.metadata)

        # These should be set on all datasets by default, but we don't enforce this rigidly
        metadata_keys = sorted([m for m in metadata_dict], key=lambda x:x.lower())

        return { k : str(metadata_dict[k]) for k in metadata_keys }


    def get_hash(self, val, metadata):

        if metadata is None:
            metadata = self.get_metadata_as_dict()

        if val is None:
            value = self.parse_value()
        else:
            value = val

        dataset_dict = {'unit_id'  : self.unit_id,
                        'type'     : self.type,
                        'value'    : value,
                        'metadata' : metadata}

        data_hash = generate_data_hash(dataset_dict)

        return data_hash
